@@ -2,7 +2,7 @@
 pygltflib : A Python library for reading, writing and handling GLTF files.
 
 
-Copyright (c) 2018, 2019 Luke Miller
+Copyright (c) 2018,2021 Luke Miller
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -35,6 +35,7 @@ from enum import Enum
 import json
 import mimetypes
 from pathlib import Path
+from shutil import copyfile
 from typing import Any, Dict, List
 from typing import Callable, Optional, Tuple, TypeVar, Union
 import struct
@@ -45,7 +46,7 @@ from dataclasses_json.core import _decode_dataclass
 from dataclasses_json.core import _ExtendedEncoder as JsonEncoder
 from deprecated import deprecated
 
-__version__ = "1.13.9"
+__version__ = "1.14.4"
 
 """
 About the GLTF2 file format:
@@ -197,7 +198,9 @@ def delete_empty_keys(dictionary):
     for key, value in list(dictionary.items()):
         if value is None or (hasattr(value, '__iter__') and len(value) == 0):
             del dictionary[key]
-        elif isinstance(value, dict):
+        elif isinstance(value, dict) and key != "extensions":
+            # delete empty dicts except when the dictionary is an extension inside "extensions".
+            # The extension exemption is because we use dicts for extensions instead of dataclass objects
             delete_empty_keys(value)
         elif isinstance(value, list):
             for item in value:
@@ -298,7 +301,7 @@ class Attributes:
 @dataclass_json
 @dataclass
 class Primitive(Property):
-    attributes: Attributes = Attributes()  # required
+    attributes: Attributes = field(default_factory=Attributes)  # required
     indices: Optional[int] = None
     mode: Optional[int] = TRIANGLES
     material: Optional[int] = None
@@ -572,8 +575,9 @@ class GLTF2(Property):
     scenes: List[Scene] = field(default_factory=list)
     skins: List[Skin] = field(default_factory=list)
     textures: List[Texture] = field(default_factory=list)
-#    _glb_data: Any = None
-#    _path: Any = None
+
+    #    _glb_data: Any = None
+    #    _path: Any = None
 
     def binary_blob(self):
         """ Get the binary blob associated with glb files if available
@@ -583,9 +587,12 @@ class GLTF2(Property):
         """
         return getattr(self, "_glb_data", None)
 
+    def set_binary_blob(self, blob):
+        setattr(self, "_glb_data", blob)
+
     def destroy_binary_blob(self):
         if hasattr(self, "_glb_data"):
-            self._glb_data = None
+            setattr(self, "_glb_data", None)
 
     def load_file_uri(self, uri):
         """
@@ -629,6 +636,7 @@ class GLTF2(Property):
                           "Please open an issue at https://gitlab.com/dodgyville/pygltflib/issues")
         return uri_format
 
+    # noinspection PyPep8Naming
     def remove_bufferView(self, buffer_view_id):
         """
         Remove a bufferView and update all the bufferView pointers in the GLTF object
@@ -656,16 +664,79 @@ class GLTF2(Property):
 
         return bufferView
 
-    def export_image_to_file(self, image_index, override=False):
+
+    def export_datauri_as_image_file(self, data_uri, name, destination, override=False, index=0):
+        """ convert data uri to image file
+            If destination is full path and file name, use that.
+            If destination is just a directory, use the name of the data_uri
+        """
+        header, encoded = data_uri.split(",", 1)
+        mime = header.split(":")[1].split(";")[0]
+        if name:  # use image.name
+            file_name = name
+        else:
+            extension = mimetypes.guess_extension(mime)
+            file_name = f"{index}{extension}"
+        destination = Path(destination)
+        if destination.is_dir():
+            image_path = destination / file_name
+        else:  # assume filepath
+            image_path = destination
+        if image_path.is_file() and not override:
+            warnings.warn(f"Unable to write image file, a file already exists at {image_path}")
+            return None
+        data = base64.b64decode(encoded)
+
+        with open(image_path, "wb") as image_file:
+            image_file.write(data)
+        return file_name
+
+    def export_fileuri_as_image_file(self, file_uri, destination, override=False):
+        """ Export file uri as another image file (ie copy out of GLTF into own location) """
+        path = getattr(self, "_path", Path())
+        image_path = Path(path / file_uri)
+        if not image_path.exists():
+            warnings.warn(f"Unable to find image {image_path} for export.")
+            return None
+        if image_path.is_file() and not override:
+            warnings.warn(f"Unable to write image file, a file already exists at {image_path}")
+            return None
+
+        copyfile(image_path, destination)
+        return file_uri
+
+    def export_image(self, image_index, destination='', override=False):
+        """ Directly export an image to a file without affecting GLTF """
+        destination = Path(destination)
+        image = self.images[image_index]
+        if image.uri and not image.uri.startswith('data:'):  # copy file to new location
+            self.export_fileuri_as_image_file(image.uri, destination)
+        elif image.bufferView is not None:
+            warnings.warn("pygltflib.export_image currently unable to convert image stored buffers to image file. "
+                          "Try GLTF.convert_images() first. "
+                          "Please open an issue at https://gitlab.com/dodgyville/pygltflib/issues")
+
+        elif image.uri.startswith('data:'):
+            file_name = self.export_datauri_as_image_file(image.uri, image.name, destination, override, image_index)
+            return file_name
+
+    def export_image_to_file(self, image_index, destination_path='', override=False):
+        """
+        Used primarly by convert_images. To export images consider using GLTF2.export_image
+
+        image_index (int): Image index
+        destination_path (str|Path): Path where to save images. Images will also be loaded from this path if needed.
+        override (bool): Only save image if it does not already exist
+        """
+        destination_path = Path(destination_path)
         image = self.images[image_index]
-        path: Path = getattr(self, "_path", Path())
         if image.uri and not image.uri.startswith('data:'):
             # already in file format
-            if not (path / image.uri).exists():
-                warnings.warn(f"Image {image_index} is already stored in a file {path / image.uri} but file"
+            if not (destination_path / image.uri).exists():
+                warnings.warn(f"Image {image_index} is already stored in a file {destination_path / image.uri} but file"
                               f"does not appear to exist.")
             return None
-        elif image.bufferView:
+        elif image.bufferView is not None:
             # TODO: remove bufferView from GLTF when create images or datauris from buffer data
             bufferView = self.bufferViews[image.bufferView]
             buffer = self.buffers[bufferView.buffer]
@@ -679,57 +750,51 @@ class GLTF2(Property):
                 data = self.binary_blob()
                 extension = mimetypes.guess_extension(image.mimeType)
                 file_name = f"{image_index}{extension}"
-                with open(file_name, "wb") as f:
+                image_path = destination_path / file_name
+                if image_path.is_file() and not override:
+                    warnings.warn(f"Unable to write image file, a file already exists at {image_path}")
+                    return None
+
+                with open(image_path, "wb") as f:
                     f.write(data[bufferView.byteOffset:bufferView.byteOffset + bufferView.byteLength])
                 return file_name
             return None
         elif image.uri.startswith('data:'):
-            #  convert data uri to image file
-            header, encoded = image.uri.split(",", 1)
-            mime = header.split(":")[1].split(";")[0]
-            if image.name:  # use image.name
-                file_name = image.name
-            else:
-                extension = mimetypes.guess_extension(mime)
-                file_name = f"{image_index}{extension}"
-            image_path = path / file_name
-            if image_path.is_file() and not override:
-                warnings.warn(f"Unable to write image file, a file already exists at {image_path}")
-                return None
-            data = base64.b64decode(encoded)
-
-            with open(image_path, "wb") as image_file:
-                image_file.write(data)
+            file_name = self.export_datauri_as_image_file(image.uri, image.name, destination_path, override, image_index)
             return file_name
 
-    def convert_images(self, image_format, override=False):
+    def convert_images(self, image_format, path=None, override=False):
         """
         GLTF files can store the image data in three different formats: In the buffers, as a data
         uri string and as external images files. This converts the images between the formats.
 
-        image_format (ImageFormat.ENUM)
+        image_format (ImageFormat.ENUM): Destination format to convert images
+        path (str|Path): Path to the directory to use for loading or saving images
         override (bool): Override an image file if it already exists and is about to be replaced
 
         """
-        path: Path = getattr(self, "_path", Path())
+        if path is None:
+            path = getattr(self, "_path", Path())
+        else:
+            path = Path(path)
         for image_index, image in enumerate(self.images):
             if image_format == ImageFormat.DATAURI:  # convert to data uri
                 # load an image file or pull from the buffer
                 if image.uri:  # either already in a datauri format, or in a file
                     if not image.uri.startswith('data:'):  # not in data format, so assume a file name
                         # data is stored in a file, so load into data uri
-                        image_path = path / image.uri
+                        image_path = Path(path / image.uri)
                         if not image_path.exists():
                             warnings.warn(f"Expected image file at {image_path} not found.")
                             continue
                         mime: str
-                        mime, _ = mimetypes.guess_type(image_path)
+                        mime, _ = mimetypes.guess_type(str(image_path))
 
                         with open(image_path, "rb") as image_file:
                             encoded_string = str(base64.b64encode(image_file.read()).decode('utf-8'))
                             image.name = copy.copy(image.uri) if not image.name else image.name
                             image.uri = f'data:{mime};base64,{encoded_string}'
-                elif image.bufferView:
+                elif image.bufferView is not None:
                     # TODO: remove bufferView from GLTF when create images or datauris from buffer data
                     warnings.warn("pygltflib currently does not remove image data "
                                   "from the buffer when converting to data uri."
@@ -744,14 +809,14 @@ class GLTF2(Property):
 
             elif image_format == ImageFormat.BUFFERVIEW:  # convert to buffer
                 #  load image data into the buffer and and add a bufferView
-                if image.bufferView:
+                if image.bufferView is not None:
                     # already in bufferview format
                     continue
                 warnings.warn("pygltflib currently unable to add image data to buffers."
                               "Please open an issue at https://gitlab.com/dodgyville/pygltflib/issues")
                 continue
-            elif image_format == ImageFormat.FILE:   # convert to images
-                file_name = self.export_image_to_file(image_index, override)
+            elif image_format == ImageFormat.FILE:  # convert to images
+                file_name = self.export_image_to_file(image_index, path, override)
                 if file_name:  # replace data uri with pointer to file
                     image.uri = file_name
 
@@ -787,7 +852,7 @@ class GLTF2(Property):
                     warnings.warn("pygltflib currently unable to convert multiple buffers to a single binary blob."
                                   "Please open an issue at https://gitlab.com/dodgyville/pygltflib/issues")
                     return
-                self._glb_data = data
+                self.set_binary_blob(data)
                 buffer.uri = ''
             elif buffer_format == BufferFormat.DATAURI:
                 # convert buffer
@@ -881,75 +946,77 @@ class GLTF2(Property):
         self.buffers = original_buffers  # restore buffers
         return True
 
+    def save_to_bytes(self):
+        # setup
+        buffer_blob = b''
+        original_buffer_views = copy.deepcopy(self.bufferViews)
+        original_buffers = copy.deepcopy(self.buffers)
+
+        offset = 0
+        new_buffer = Buffer()
+        path = getattr(self, "_path", Path())
+        for i, bufferView in enumerate(self.bufferViews):
+            buffer = self.buffers[bufferView.buffer]
+            if buffer.uri == '':  # assume loaded from glb binary file
+                data = self.binary_blob()
+            elif buffer.uri.startswith("data"):
+                data = self.decode_data_uri(buffer.uri)
+            elif Path(path, buffer.uri).is_file():
+                with open(Path(path, buffer.uri), 'rb') as fb:
+                    data = fb.read()
+            else:
+                warnings.warn(f"Unable to save bufferView {buffer.uri[:20]} to glb, skipping. "
+                              "Please open an issue at https://gitlab.com/dodgyville/pygltflib/issues")
+                continue
+            byte_offset = bufferView.byteOffset if bufferView.byteOffset is not None else 0
+            byte_length = bufferView.byteLength
+            if byte_length % 4 != 0:  # pad each segment of binary blob
+                byte_length += 4 - byte_length % 4
+
+            buffer_blob += data[byte_offset:byte_offset + byte_length]
+
+            bufferView.byteOffset = offset
+            bufferView.byteLength = byte_length
+            bufferView.buffer = 0
+            offset += byte_length
+
+        new_buffer.byteLength = len(buffer_blob)
+        self.buffers = [new_buffer]
+
+        json_blob = self.gltf_to_json().encode("utf-8")
+
+        # pad each blob if needed
+        if len(json_blob) % 4 != 0:
+            json_blob += b'   '[0:4 - len(json_blob) % 4]
+
+        version = struct.pack('<I', GLTF_VERSION)
+        chunk_header_len = 8
+        length = len(MAGIC) + len(version) + 4 + chunk_header_len * 2 + len(json_blob) + len(buffer_blob)
+        self.bufferViews = original_buffer_views  # restore unpacked bufferViews
+        self.buffers = original_buffers  # restore unpacked buffers
+
+        # header is MAGIC, version, length
+        # json chunk is json_blob length, JSON, json_blob
+        # buffer chunk is length of buffer_blob, utf-8, buffer_blob
+        return [
+            MAGIC,
+            version,
+            struct.pack('<I', length),
+            struct.pack('<I', len(json_blob)),
+            bytes(JSON, 'utf-8'),
+            json_blob,
+            struct.pack('<I', len(buffer_blob)),
+            bytes(BIN, 'utf-8'),
+            buffer_blob
+        ]
+
     def save_binary(self, fname):
         with open(fname, 'wb') as f:
-            # setup
-
-            buffer_blob = b''
-            original_buffer_views = copy.deepcopy(self.bufferViews)
-            original_buffers = copy.deepcopy(self.buffers)
-
-            offset = 0
-            new_buffer = Buffer()
-            path = getattr(self, "_path", Path())
-            for i, bufferView in enumerate(self.bufferViews):
-                buffer = self.buffers[bufferView.buffer]
-                if buffer.uri == '':  # assume loaded from glb binary file
-                    data = self.binary_blob()
-                elif buffer.uri.startswith("data"):
-                    warnings.warn(f"Unable to save data uri bufferView {buffer.uri[:20]} to glb, "
-                                  "please save in gltf format instead or use the convert_buffers method first."
-                                  "Please open an issue at https://gitlab.com/dodgyville/pygltflib/issues")
-                    return False
-                elif Path(path, buffer.uri).is_file():
-                    with open(Path(path, buffer.uri), 'rb') as fb:
-                        data = fb.read()
-                else:
-                    warnings.warn(f"Unable to save bufferView {buffer.uri[:20]} to glb, skipping. "
-                                  "Please open an issue at https://gitlab.com/dodgyville/pygltflib/issues")
-                    continue
-                byte_offset = bufferView.byteOffset if bufferView.byteOffset is not None else 0
-                byte_length = bufferView.byteLength
-                if byte_length % 4 != 0:  # pad each segment of binary blob
-                    byte_length += 4 - byte_length % 4
-
-                buffer_blob += data[byte_offset:byte_offset + byte_length]
-
-                bufferView.byteOffset = offset
-                bufferView.byteLength = byte_length
-                bufferView.buffer = 0
-                offset += byte_length
-
-            new_buffer.byteLength = len(buffer_blob)
-            self.buffers = [new_buffer]
-
-            json_blob = self.gltf_to_json().encode("utf-8")
-
-            # pad each blob if needed
-            if len(json_blob) % 4 != 0:
-                json_blob += b'   '[0:4 - len(json_blob) % 4]
-
-            version = struct.pack('<I', GLTF_VERSION)
-            chunk_header_len = 8
-            length = len(MAGIC) + len(version) + 4 + chunk_header_len * 2 + len(json_blob) + len(buffer_blob)
-            self.bufferViews = original_buffer_views  # restore unpacked bufferViews
-            self.buffers = original_buffers  # restore unpacked buffers
-
-            # header
-            f.write(MAGIC)
-            f.write(version)
-            f.write(struct.pack('<I', length))
-
-            # json chunk
-            f.write(struct.pack('<I', len(json_blob)))
-            f.write(bytes(JSON, 'utf-8'))
-            f.write(json_blob)
-
-            # buffer chunk
-            f.write(struct.pack('<I', len(buffer_blob)))
-            f.write(bytes(BIN, 'utf-8'))
-            f.write(buffer_blob)
-
+            glb_structure = self.save_to_bytes()
+            if not glb_structure:
+                return False
+            for data in glb_structure:
+                f.write(data)
         return True
 
     def save(self, fname, asset=Asset()):
@@ -1002,7 +1069,7 @@ class GLTF2(Property):
                 raw_json = data[index:index + chunk_length].decode("utf-8")
                 obj = cls.from_json(raw_json, infer_missing=True)
             else:
-                obj._glb_data = data[index:index + chunk_length]
+                obj.set_binary_blob(data[index:index + chunk_length])
             index += chunk_length
             i += 1
         return obj
@@ -1021,9 +1088,6 @@ class GLTF2(Property):
     @classmethod
     def load(cls, fname):
         path = Path(fname)
-        if not path.is_file():
-            print("ERROR: File not found", fname)
-            return None
         ext = path.suffix
         if ext.lower() in [".bin", ".glb"]:
             obj = cls.load_binary(fname)
